from typing import List, NamedTuple, Optional, Union, cast

from autobot import api
from autobot.transforms import TransformType


class Prompt(NamedTuple):
    text: str
    max_tokens: int
    stop: Optional[Union[str, List[str]]]


def make_prompt(
    snippet: str,
    *,
    transform_type: TransformType,
    before_text: str,
    after_text: str,
    before_description: str,
    after_description: str,
) -> Prompt:
    """Construct a Prompt object from a source snippet."""
    node_name = transform_type.plaintext_name()
    return Prompt(
        f"""### Python {node_name} {before_description}
{before_text}

### The same Python {node_name} {after_description}
{after_text}

### Python {node_name} {before_description}
{snippet}
### End of {node_name}

### Now rewrite the Python {node_name} {after_description}
""",
        max_tokens=len(snippet) // 2,
        stop=f"### End of {node_name}",
    )


def resolve_prompt(prompt: Prompt) -> str:
    """Generate a completion for a prompt."""
    response = api.create_completion(
        prompt=prompt.text,
        max_tokens=prompt.max_tokens,
        stop=prompt.stop,
        temperature=0,
        model="text-davinci-002",
    )
    for choice in response["choices"]:
        return cast(str, choice["text"])
    else:
        raise Exception("Request failed to generate choices.")



import ast
import re
from typing import Generator, List, NamedTuple, Type, Union


def decontextualize(source: str, node: ast.stmt) -> "Snippet":
    """Decontextualize a snippet from its originating source code.

    Takes the originating source code as input, along with the node to decontextualize,
    and extracts the code as a snippet, removing any indentation.
    """
    # Extract the source segment.
    source_segment = ast.get_source_segment(source, node, padded=True)
    assert source_segment, "Unable to find source segment."

    # Dedent the code:.
    # TODO(charlie): This isn't safe. For example, there could be multi-line strings
    # within a function that need this padding.
    lines = source_segment.splitlines()
    if m := re.match(r"(\s+)", lines[0]):
        padding = m.group()
        source_segment = "\n".join([line.removeprefix(padding) for line in lines])
    else:
        padding = ""

    return Snippet(source_segment, padding, node.lineno)


def recontextualize(snippet: "Snippet", source: str) -> List[str]:
    """Recontextualize a snippet within its originating source code.

    Takes the originating source code and snippet as input, and outputs the lines of the
    source code up to and including the snippet, with the snippet adjusted to match the
    indentation of its originating context.
    """
    lines: List[str] = []

    # Prepend any lines of the originating source code that precede the snippet.
    source_lines = source.splitlines()
    if snippet.lineno > 1:
        for i in range(snippet.lineno - 1):
            lines.append(source_lines[i])

    # Tack on the snippet itself, with indentation re-applied.
    for line in snippet.text.splitlines():
        lines.append(snippet.padding + line)

    return lines


class Snippet(NamedTuple):
    """A snippet extracted from source code."""

    text: str
    padding: str
    lineno: int

    @classmethod
    def from_node(cls, source: str, node: ast.stmt) -> "Snippet":
        return decontextualize(source, node)


class ClassDefVisitor(ast.NodeVisitor):
    """NodeVisitor to collect all ClassDef nodes in an AST."""

    def __init__(self) -> None:
        self.nodes: List[ast.ClassDef] = []

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        self.nodes.append(node)


class FunctionDefVisitor(ast.NodeVisitor):
    """NodeVisitor to collect all FunctionDef nodes in an AST."""

    def __init__(self) -> None:
        self.nodes: List[ast.FunctionDef] = []

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self.nodes.append(node)


def iter_snippets(
    source: str,
    node_type: Type[ast.stmt],
) -> Generator[Snippet, None, None]:
    """Generate all snippets from the provided source code.

    Returns: a tuple of (text to fix, any indentation that was removed from the
        snippet, line number in the source file).
    """
    visitor: Union[ClassDefVisitor, FunctionDefVisitor]
    if node_type == ast.ClassDef:
        visitor = ClassDefVisitor()
    elif node_type == ast.FunctionDef:
        visitor = FunctionDefVisitor()
    else:
        raise NotImplementedError(f"Unhandled AST node type: {node_type}")

    visitor.visit(ast.parse(source))

    for node in visitor.nodes:
        yield Snippet.from_node(source, node)



"""Interface to the OpenAI API."""
import hashlib
import json
import logging
import os
from typing import List, Optional, Union

import openai

from autobot.utils import cache


def init() -> None:
    openai.organization = os.environ["OPENAI_ORGANIZATION"]
    openai.api_key = os.environ["OPENAI_API_KEY"]


def create_completion(
    prompt: str,
    max_tokens: int,
    temperature: int = 0,
    model: str = "text-davinci-002",
    stop: Optional[Union[str, List[str]]] = None,
) -> openai.Completion:
    request_hash = hashlib.md5(
        json.dumps(
            {
                "prompt": prompt,
                "max_tokens": max_tokens,
                "temperature": temperature,
                "model": model,
                "stop": stop,
            }
        ).encode("utf-8")
    ).hexdigest()

    if response := cache.get_from_cache(request_hash):
        logging.debug("Reading response from cache...")
        return response

    response = openai.Completion.create(
        model=model,
        prompt=prompt,
        temperature=temperature,
        max_tokens=max_tokens,
        stop=stop,
    )
    cache.set_in_cache(request_hash, response)
    return response



import enum
import os
from typing import Dict, List, Optional

from colorama import Fore
from rich.console import Console

from autobot.constants import PATCH_DIR
from autobot.refactor import patch
from autobot.utils.getch import getch


class Resolution(enum.Enum):
    ACCEPT = "a"
    REJECT = "r"
    SKIP = "s"

    @classmethod
    def from_code(cls, code: str) -> Optional["Resolution"]:
        try:
            return cls(code)
        except ValueError:
            return None


def run_review() -> None:
    patches: List[str] = []
    for root, _, filenames in os.walk(PATCH_DIR):
        for filename in filenames:
            if filename.endswith(".patch"):
                patches.append(os.path.join(root, filename))

    console = Console()

    patches_by_resolution: Dict[Resolution, List[str]] = {
        Resolution.ACCEPT: [],
        Resolution.REJECT: [],
        Resolution.SKIP: [],
    }
    num_patches = len(patches)
    for i, patch_file in enumerate(patches):
        if patch.can_apply(patch_file):
            with console.screen(hide_cursor=False):
                with open(patch_file, "r") as fp:
                    contents = fp.read()

                console.print(
                    f"[bold][white]Reviewing [[yellow]{i + 1}/{num_patches}[/yellow]]"
                )
                console.print(f"Patch file: [cyan]{os.path.basename(patch_file)}")

                console.print()
                for line in contents.splitlines():
                    stripped = line.strip()
                    if len(stripped) == 0:
                        line = stripped
                    if line.startswith("-"):
                        print(f"{Fore.RED}{line}{Fore.RESET}")
                    elif line.startswith("+"):
                        print(f"{Fore.GREEN}{line}{Fore.RESET}")
                    else:
                        print(line)
                console.print()

                console.print("  [bold green]a[/] accept  [grey46]apply the patch[/]")
                console.print("  [bold red]r[/] reject  [grey46]reject the patch[/]")
                console.print("  [bold yellow]s[/] skip    [grey46]skip the patch[/]")

                try:
                    while (resolution := Resolution(getch())) is None:
                        pass
                except KeyboardInterrupt:
                    exit(0)

                patches_by_resolution[resolution].append(patch_file)
                if resolution == Resolution.ACCEPT:
                    # Apply the patch.
                    patch.apply(patch_file)
                    os.remove(patch_file)
                elif resolution == Resolution.REJECT:
                    # Reject the patch.
                    os.remove(patch_file)
                elif resolution == Resolution.SKIP:
                    # Do nothing.
                    pass
                else:
                    raise ValueError(f"Unexpected resolution: {resolution}")

    if num_patches > 0:
        if num_patches == 1:
            console.print(f"[bold]Done![/] Reviewed {len(patches)} patch.")
        else:
            console.print(f"[bold]Done![/] Reviewed {len(patches)} patches.")
        for resolution in patches_by_resolution:
            if patches_by_resolution[resolution]:
                if resolution == Resolution.ACCEPT:
                    console.print("[green]Accepted:")
                elif resolution == Resolution.REJECT:
                    console.print("[red]Rejected:")
                elif resolution == Resolution.SKIP:
                    console.print("[yellow]Skipped:")
                else:
                    raise ValueError(f"Unexpected resolution: {resolution}")

                for patch_file in patches_by_resolution[resolution]:
                    print(f"  {os.path.relpath(patch_file, PATCH_DIR)}")
    else:
        console.print("[bold]Done![/] No patches to review.")



import difflib
import functools
import logging
import os.path
from multiprocessing.pool import ThreadPool
from typing import Dict, List, Set

from colorama import Fore
from rich.console import Console

from autobot import prompt
from autobot.constants import PATCH_DIR
from autobot.snippet import Snippet, iter_snippets, recontextualize
from autobot.transforms import TransformType
from autobot.utils import filesystem


def fix_text(
    text: str,
    *,
    transform_type: TransformType,
    before_text: str,
    after_text: str,
    before_description: str,
    after_description: str,
) -> tuple[str, str]:
    """Generate a fix for a snippet.

    Returns: a tuple of (snippet, suggested fix), to play nicely with multiprocessing.
    """
    return text, prompt.resolve_prompt(
        prompt.make_prompt(
            text,
            transform_type=transform_type,
            before_text=before_text,
            after_text=after_text,
            before_description=before_description,
            after_description=after_description,
        )
    )


def run_refactor(
    *,
    title: str,
    before_filename: str,
    after_filename: str,
    targets: List[str],
    before_description: str,
    after_description: str,
    transform_type: TransformType,
    nthreads: int,
) -> None:
    with open(before_filename, "r") as fp:
        before_text = fp.read()

    with open(after_filename, "r") as fp:
        after_text = fp.read()

    console = Console()

    console.print(f"[bold]Running [bold cyan]{title}[/] based on user-provided example")

    console.print("-" * len(f"Running {title} based on user-provided example"))
    for line in difflib.unified_diff(
        before_text.splitlines(),
        after_text.splitlines(),
        lineterm="",
        fromfile=os.path.join("a", before_filename),
        tofile=os.path.join("b", after_filename),
    ):
        # TODO(charlie): Why is this necessary? Without it, blank lines contain
        # a single space.
        if len(line.strip()) == 0:
            line = line.strip()

        if line.startswith("-"):
            print(f"{Fore.RED}{line}{Fore.RESET}")
        elif line.startswith("+"):
            print(f"{Fore.GREEN}{line}{Fore.RESET}")
        else:
            print(line)
    console.print("-" * len(f"Running {title} based on user-provided example"))
    console.print()

    console.print("[bold]1. Collecting Python files...")
    targets = filesystem.collect_python_files(targets)

    # Deduplicate targets, such that if we need to apply the same fix to a bunch of
    # snippets, we only make a single API call.
    console.print("[bold]2. Extracting AST nodes...")
    filename_to_snippets = Dict[str, List[Snippet]]()
    all_snippet_texts = Set[str]()
    for filename in targets:
        with open(filename, "r") as fp:
            source_code = fp.read()

        filename_to_snippets[filename] = []
        for snippet in iter_snippets(source_code, transform_type.ast_node_type()):
            if len(snippet.text) > 1600:
                logging.warning(
                    f"Snippet at {filename}:{snippet.lineno} is too long "
                    f"({len(snippet.text)} > 1600); skipping..."
                )
                continue
            filename_to_snippets[filename].append(snippet)
            all_snippet_texts.add(snippet.text)

    # Map from snippet text to suggested fix.
    console.print("[bold]3. Generating LLM completions...")
    snippet_text_to_completion = Dict[str, str]()
    with ThreadPool(processes=nthreads) as pool:
        for text, completion in pool.map(
            functools.partial(
                fix_text,
                transform_type=transform_type,
                before_text=before_text,
                after_text=after_text,
                before_description=before_description,
                after_description=after_description,
            ),
            all_snippet_texts,
        ):
            snippet_text_to_completion[text] = completion

    # Format each suggestion as a patch.
    console.print("[bold]4. Constructing Git patches...")
    count: int = 0
    for target in filename_to_snippets:
        with open(target, "r") as fp:
            source = fp.read()

        for text, padding, lineno in filename_to_snippets[target]:
            before_text = text
            after_text = snippet_text_to_completion[text]
            patch: str = ""
            for line in difflib.unified_diff(
                recontextualize(Snippet(before_text, padding, lineno), source),
                recontextualize(Snippet(after_text, padding, lineno), source),
                lineterm="",
                fromfile=os.path.join("a", target),
                tofile=os.path.join("b", target),
            ):
                # TODO(charlie): Why is this necessary? Without it, blank lines contain
                # a single space.
                stripped = line.strip()
                if len(stripped) == 0:
                    line = stripped

                patch += line
                patch += "\n"

            if patch:
                # Save the patch.
                (target_filename, _) = os.path.splitext(target)
                patch_filename = os.path.join(
                    PATCH_DIR,
                    f"{target_filename}-{lineno}.patch",
                )
                os.makedirs(os.path.dirname(patch_filename), exist_ok=True)
                with open(patch_filename, "w") as fp:
                    fp.write(patch)

                count += 1

    console.print()
    if count == 0:
        console.print(f"[bold white]✅ Done! No suggestions found.")
    elif count == 1:
        console.print(f"[bold white]✨ Done! Generated {count} patch.")
    else:
        console.print(f"[bold white]✨ Done! Generated {count} patches.")



"""Entrypoint to the autobot CLI."""
import argparse
import json
import logging
import os
from typing import Any, List

from dotenv import load_dotenv
from rich.console import Console

from autobot.transforms import TransformType


def run(options: Any) -> None:
    from autobot import api
    from autobot.refactor import run_refactor

    api.init()

    schematic: str = options.schematic
    nthreads: int = options.nthreads
    verbose: bool = options.verbose

    logging.basicConfig(
        format="%(asctime)s %(levelname)-8s %(message)s",
        datefmt="%m-%d %H:%M:%S",
        level=logging.DEBUG if verbose else logging.INFO,
    )

    console = Console()

    # Extract various inputs based on the schematic location.
    if not os.path.isdir(schematic):
        console.print(f"[bold red]error[/]  Directory not found: {schematic}")
        exit(1)

    before_filename: str = os.path.join(schematic, "before.py")
    if not os.path.isfile(before_filename):
        console.print(f"[bold red]error[/]  Unable to find file: {before_filename}")
        exit(1)

    after_filename: str = os.path.join(schematic, "after.py")
    if not os.path.isfile(after_filename):
        console.print(f"[bold red]error[/]  Unable to find file: {after_filename}")
        exit(1)

    autobot: str = os.path.join(schematic, "autobot.json")
    if not os.path.isfile(autobot):
        console.print(f"[bold red]error[/]  Unable to find file: {autobot}")
        exit(1)

    with open(autobot, "r") as fp:
        metadata = json.load(fp)
        if not (before_description := metadata.get("before_description")):
            console.print(
                f"[bold red]error[/]  autobot.json is missing `before_description`"
            )
            exit(1)
        if not (after_description := metadata.get("after_description")):
            console.print(
                f"[bold red]error[/]  autobot.json is missing `after_description`"
            )
            exit(1)
        if not (transform_type_raw := metadata.get("transform_type")):
            console.print(
                f"[bold red]error[/]  autobot.json is missing `transform_type`"
            )
            exit(1)

        try:
            transform_type: TransformType = TransformType(transform_type_raw)
        except ValueError:
            console.print(
                "[bold red]error[/]  "
                f"Invalid `transform_type`: '{transform_type_raw}'. Choose one of: "
                f"{[member.value for member in TransformType]}."
            )
            exit(1)

    targets: List[str] = options.target

    run_refactor(
        title=os.path.basename(schematic),
        before_filename=before_filename,
        after_filename=after_filename,
        targets=targets,
        before_description=before_description,
        after_description=after_description,
        transform_type=transform_type,
        nthreads=nthreads,
    )


def review(options: Any) -> None:
    from autobot.review import run_review

    run_review()


def main() -> None:
    load_dotenv()

    parser = argparse.ArgumentParser(prog="autobot")
    subparsers = parser.add_subparsers()

    # autobot run
    parser_run = subparsers.add_parser("run")
    parser_run.add_argument(
        "schematic", type=str, help="Path to the autobot schematic."
    )
    parser_run.add_argument(
        "target", type=str, nargs="+", help="Path to the files to refactor."
    )
    parser_run.add_argument(
        "--nthreads",
        type=int,
        default=8,
        help="The number of threads to use when generating completions.",
    )
    parser_run.add_argument(
        "--verbose",
        action="store_true",
        help="Whether to print our completions.",
    )
    parser_run.set_defaults(func=run)

    # autobot review
    parser_review = subparsers.add_parser("review")
    parser_review.set_defaults(func=review)

    args = parser.parse_args()
    if hasattr(args, "func"):
        args.func(args)
    else:
        parser.print_usage()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['autobot', 'autobot.refactor', 'autobot.review', 'autobot.utils']

package_data = \
{'': ['*']}

install_requires = \
['colorama>=0.4.5,<0.5.0',
 'openai>=0.23.0,<0.24.0',
 'python-dotenv>=0.21.0,<0.22.0',
 'rich>=12.5.1,<13.0.0']

entry_points = \
{'console_scripts': ['autobot = autobot.main:main']}

setup_kwargs = {
    'name': 'autobot-ml',
    'version': '0.0.2',
    'description': 'An automated code refactoring tool powered by GPT-3.',
    'long_description': '# autobot\n\nAn automated code refactoring tool powered by GPT-3. Like GitHub Copilot, for your existing\ncodebase.\n\nAutobot takes an example change as input and generates patches for you to review by scanning your\ncodebase for similar code blocks and "applying" that change to the existing source code.\n\n## Getting started\n\nAutobot is available as [`autobot-ml`](https://pypi.org/project/autobot-ml/) on PyPI:\n\n```shell\npip install autobot-ml\n```\n\nAutobot depends on the [OpenAI API](https://openai.com/api/) and, in particular, expects your OpenAI\norganization ID and API key to be exposed as the `OPENAI_ORGANIZATION` and `OPENAI_API_KEY`\nenvironment variables, respectively.\n\nAutobot can also read from a `.env` file:\n\n```\nOPENAI_ORGANIZATION=${YOUR_OPENAI_ORGANIZATION}\nOPENAI_API_KEY=${YOUR_OPENAI_API_KEY}\n```\n\n## Example usage\n\n_TL;DR: Autobot is a command-line tool. To generate patches, use `autobot run`; to review the\ngenerated patches, use `autobot review`._\n\nAutobot is designed around a two-step workflow.\n\nIn the first step (`autobot run {schematic} {files_to_analyze}`), we point Autobot to (1) the\n"schematic" that defines our desired change and (2) the files to which the change should be\napplied.\n\nIn the second step (`autobot review`), we review the patches that Autobot generated and,\nfor each suggested change, either apply it to the codebase or reject it.\n\nFor example: to remove any usages of NumPy\'s deprecated `np.int` and associated aliases, we\'d first\nrun `autobot run ./schematics/numpy_builtin_aliases ./path/to/main.py`, followed by\n`autobot review`.\n\n### Implementing a novel refactor\n\nEvery refactor facilitated by Autobot requires a "schematic". Autobot ships with a few example\nschematics in the `schematics` directory, but it\'s intended to be used with user-provided\nschematics.\n\nA schematic is a directory containing three files:\n\n1. `before.py`: A code snippet demonstrating the "before" state of the refactor.\n2. `after.py`: A code snippet demonstrating the "after" state of the refactor.\n3. `autobot.json`: A JSON object containing a plaintext description of the\n   before (`before_description`) and after (`after_description`) states, along with\n   the `transform_type` ("Function" or "Class").\n\nFor example: in Python 3, `class Foo(object)` is equivalent to `class Foo`. To automatically remove\nthose useless object inheritances from our codebase, we\'d create a `useless_object_inheritance`\ndirectory, and add the above files.\n\n```python\n# before.py\nclass Foo(Bar, object):\n    def __init__(self, x: int) -> None:\n        self.x = x\n```\n\n```python\n# after.py\nclass Foo(Bar):\n    def __init__(self, x: int) -> None:\n        self.x = x\n```\n\n```json\n// autobot.json\n{\n    "before_description": "with object inheritance",\n    "after_description": "without object inheritance",\n    "transform_type": "Class"\n}\n```\n\nWe\'d then run `autobot run useless_object_inheritance /path/to/file/or/directory` to generate\npatches, followed by `autobot review` to apply or reject the suggested changes.\n\n## Limitations\n\n1. To speed up execution, Autobot calls out to the OpenAI API in parallel. If you haven\'t upgraded\n   to a paid account, you may hit rate-limit errors. You can pass `--nthreads 1` to `autobot run`\n   to disable multi-threading.\n2. Depending on the transform type, Autobot will either generate a patch for every function or every\n   class. Any function or class that\'s "too long" will be GPT-3\'s maximum prompt size,\n\n## License\n\nMIT\n',
    'author': 'Charlie Marsh',
    'author_email': 'charlie.r.marsh@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/charliermarsh/autobot',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import subprocess


def can_apply(patch_file: str) -> bool:
    result = subprocess.run(
        ["git", "apply", "--check", patch_file],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    return result.returncode == 0


def apply(patch_file: str) -> None:
    subprocess.check_call(["git", "apply", patch_file])



"""Test cases to assess prompt completion.

Note that these tests rely on the OpenAI API.
"""
import unittest

from dotenv import load_dotenv

from autobot import api, prompt
from autobot.transforms import TransformType


class PromptIntegrationTest(unittest.TestCase):
    def setUp(self) -> None:
        load_dotenv()
        api.init()

    def test_useless_object_inheritance(self) -> None:
        before_text = """
class Foo(Bar, object):
    def __init__(self, x: int) -> None:
        self.x = x
"""

        after_text = """
class Foo(Bar):
    def __init__(self, x: int) -> None:
        self.x = x
"""

        snippet = """
class CreateTaskResponse(object):
    task_id: str
        """

        expected = """
class CreateTaskResponse:
    task_id: str
        """
        actual = prompt.resolve_prompt(
            prompt.make_prompt(
                snippet.strip(),
                transform_type=TransformType.CLASS,
                before_text=before_text.strip(),
                after_text=after_text.strip(),
                before_description="with object inheritance",
                after_description="without object inheritance",
            )
        )

        self.assertEqual(expected.strip(), actual.strip())

    def test_sorted_attributes(self) -> None:
        before_text = """
class MyUnsortedConstants:
    z = "hehehe"
    B = "aaa234"
    A = "zzz123"
    cab = "foo bar"
    Daaa = "banana"
"""

        after_text = """
class MyUnsortedConstants:
    A = "zzz123"
    B = "aaa234"
    Daaa = "banana"
    cab = "foo bar"
    z = "hehehe"
"""

        snippet = """
@dataclass
class Circle:
    \"\"\"A circle in an image, with all its pixel indices/radius/center.\"\"\"

    # All the row coordinates of pixels in the circle, suitable for logical
    # indexing with numpy.
    rr: np.ndarray

    # All the column coordinates of pixels in the circle, suitable for logical
    # indexing with numpy.
    cc: np.ndarray

    # Radius of the circle
    radius: int

    # Center of the circle (row, column)
    center: Tuple[int, int]
"""

        expected = """
@dataclass
class Circle:
    \"\"\"A circle in an image, with all its pixel indices/radius/center.\"\"\"

    # All the column coordinates of pixels in the circle, suitable for logical
    # indexing with numpy.
    cc: np.ndarray

    # Center of the circle (row, column)
    center: Tuple[int, int]

    # Radius of the circle
    radius: int

    # All the row coordinates of pixels in the circle, suitable for logical
    # indexing with numpy.
    rr: np.ndarray
"""
        actual = prompt.resolve_prompt(
            prompt.make_prompt(
                snippet.strip(),
                transform_type=TransformType.CLASS,
                before_text=before_text.strip(),
                after_text=after_text.strip(),
                before_description="with unsorted class attributes",
                after_description="with alphabetically sorted class attributes",
            )
        )

        self.assertEqual(expected.strip(), actual.strip())

    def test_standard_library_generics(self) -> None:
        before_text = """
def func(
    x: Optional[List[int]],
    y: Optional[int],
    z: Optional[Union[int, List[int]]],
) -> None:
    a: List[int] = []
    b: Set[int] = set()
    c: Optional[int] = 0
    d: Union[int, str] = 0
"""

        after_text = """
def func(
    x: list[int] | None,
    y: int | None,
    z: int | list[int] | None,
) -> None:
    a: list[int] = []
    b: set[int] = set()
    c: int | None = 0
    d: int | str = 0
"""

        snippet = """
def compute_squares(n: int) -> List[int]:
    x: List[int] = []
    for i in range(n):
        x.append(i * i)
    return x
"""

        expected = """
def compute_squares(n: int) -> list[int]:
    x: list[int] = []
    for i in range(n):
        x.append(i * i)
    return x
"""
        actual = prompt.resolve_prompt(
            prompt.make_prompt(
                snippet.strip(),
                transform_type=TransformType.CLASS,
                before_text=before_text.strip(),
                after_text=after_text.strip(),
                before_description="with typing module generics",
                after_description="with standard library generics",
            )
        )

        self.assertEqual(expected.strip(), actual.strip())

    def test_use_generator(self) -> None:
        before_text = """
def compute_squares(n: int) -> list[int]:
    squares: list[int] = []
    for i in range(n):
        squares.append(i * i)
    return squares
    """

        after_text = """
def compute_squares(n: int) -> Generator[int, None, None]:
    for i in range(n):
        yield i * i
    """

        snippet = """
def run(self, parameter_name: str, value: pd.DataFrame) -> list[SpecViolation]:
    violations: list[SpecViolation] = []
    cols = value.columns if isinstance(self.columns, _All) else self.columns
    for col in cols:
        if col not in value.columns:
            violations.append(
                SpecViolation(
                    parameter_name,
                    f"column {col} missing from dataframe",
                    pd.DataFrame,
                    value,
                )
            )
        elif value[col].isnull().values.any():
            violations.append(
                SpecViolation(
                    parameter_name,
                    f"column {col} had null values",
                    pd.DataFrame,
                    value,
                )
            )
    return violations
            """

        expected = """
def run(self, parameter_name: str, value: pd.DataFrame) -> Generator[SpecViolation, None, None]:
    cols = value.columns if isinstance(self.columns, _All) else self.columns
    for col in cols:
        if col not in value.columns:
            yield SpecViolation(
                parameter_name,
                f"column {col} missing from dataframe",
                pd.DataFrame,
                value,
            )
        elif value[col].isnull().values.any():
            yield SpecViolation(
                parameter_name,
                f"column {col} had null values",
                pd.DataFrame,
                value,
            )
            """

        actual = prompt.resolve_prompt(
            prompt.make_prompt(
                snippet.strip(),
                transform_type=TransformType.FUNCTION,
                before_text=before_text.strip(),
                after_text=after_text.strip(),
                before_description="as a list builder",
                after_description="as a generator",
            )
        )

        self.assertEqual(expected.strip(), actual.strip())



import sys
import termios
import tty


def getch() -> str:
    """Get a single character from standard input."""
    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setcbreak(sys.stdin.fileno())
        ch = sys.stdin.read(1)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
    return ch



import ast
import enum
from typing import Type


class TransformType(enum.Enum):
    CLASS = "Class"
    FUNCTION = "Function"

    def plaintext_name(self) -> str:
        if self == TransformType.CLASS:
            return "class"
        if self == TransformType.FUNCTION:
            return "function"
        raise NotImplementedError(f"Unhandled transform type: {self}")

    def ast_node_type(self) -> Type[ast.stmt]:
        if self == TransformType.CLASS:
            return ast.ClassDef
        if self == TransformType.FUNCTION:
            return ast.FunctionDef
        raise NotImplementedError(f"Unhandled transform type: {self}")



"""Filesystem-based cache for storing JSON objects."""
import json
import os
from typing import Optional, TypeVar, cast

CACHE_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.dirname(__file__))), ".autobot", "cache"
)

T = TypeVar("T")


def cache_filename(key: str) -> str:
    return os.path.join(CACHE_DIR, key)


def has_in_cache(key: str) -> bool:
    os.makedirs(os.path.dirname(cache_filename(key)), exist_ok=True)
    return os.path.exists(cache_filename(key))


def get_from_cache(key: str) -> Optional[T]:
    os.makedirs(os.path.dirname(cache_filename(key)), exist_ok=True)
    try:
        with open(cache_filename(key), "r") as fp:
            return cast(T, json.load(fp))
    except FileNotFoundError:
        return None


def set_in_cache(key: str, value: T) -> None:
    os.makedirs(os.path.dirname(cache_filename(key)), exist_ok=True)
    with open(cache_filename(key), "w") as fp:
        json.dump(value, fp)


def delete_from_cache(key: str) -> bool:
    os.makedirs(os.path.dirname(cache_filename(key)), exist_ok=True)
    try:
        os.remove(cache_filename(key))
        return True
    except FileNotFoundError:
        return False



import glob
import os
from typing import List, Set


def is_python_file(filename: str) -> bool:
    """Return True if a file appears to contain Python source code."""
    return filename.endswith(".py") or filename.endswith(".pyi")


def collect_python_files(targets: List[str]) -> List[str]:
    """Enumerate all Python files in a target."""
    collected: Set[str] = set()
    for target in targets:
        for file_or_directory in glob.iglob(target):
            if os.path.isdir(file_or_directory):
                for root, dirnames, filenames in os.walk(file_or_directory):
                    for filename in filenames:
                        if is_python_file(filename):
                            collected.add(os.path.join(root, filename))
            elif os.path.isfile(file_or_directory):
                if is_python_file(file_or_directory):
                    collected.add(file_or_directory)

    return sorted(collected)
